# -*- coding: utf-8 -*-

import os

pth = os.path.dirname(__file__)

nk = os.path.join(pth, "nk.yml")
bh = os.path.join(pth, "bh.yml")
dsge = os.path.join(pth, "dsge.yml")
ghls = os.path.join(pth, "ghls.yml")
hank = os.path.join(pth, "hank_with_comments.yml")
hank_labor = os.path.join(pth, "hank_labor.yml")
hank2 = os.path.join(pth, "hank2.yml")
hank2_no_capital = os.path.join(pth, "hank2_no_capital.yml")
tank = os.path.join(pth, "tank.yml")
